@@ -22,20 +22,37 @@ The main methods are:
 """
 
 import logging
-import multiprocessing.pool
 import os
 import subprocess
 import sys
+import requests
+if sys.version_info[0] == 2:
+    import httplib
+elif sys.version_info[0] == 3:
+    import http.client as httplib
 
 from boto.compat import BytesIO, urlsplit, six
 import boto.s3.key
 
 logger = logging.getLogger(__name__)
 
+# Multiprocessing is unavailable in App Engine (and possibly other sandboxes).
+# The only method currently relying on it is s3_iter_bucket, which is instructed
+# not to use it by the NO_MULTIPROCESSING flag.
+try:
+    import multiprocessing.pool
+except ImportError:
+    logger.warning("multiprocessing could not be imported and won't be used")
+    NO_MULTIPROCESSING = True
+    from itertools import imap
+else:
+    NO_MULTIPROCESSING = False
+
 S3_MIN_PART_SIZE = 50 * 1024**2  # minimum part size for S3 multipart uploads
+WEBHDFS_MIN_PART_SIZE = 50 * 1024**2  # minimum part size for HDFS multipart uploads
 
 
-def smart_open(uri, mode="rb"):
+def smart_open(uri, mode="rb", **kw):
     """
     Open the given S3 / HDFS / filesystem file pointed to by `uri` for reading or writing.
 
@@ -46,11 +63,12 @@ def smart_open(uri, mode="rb"):
 
     The `uri` can be either:
 
-    1. local filesystem (compressed ``.gz`` or ``.bz2`` files handled automatically):
+    1. a URI for the local filesystem (compressed ``.gz`` or ``.bz2`` files handled automatically):
        `./lines.txt`, `/home/joe/lines.txt.gz`, `file:///home/joe/lines.txt.bz2`
-    2. Amazon's S3 (can also supply credentials inside the URI):
+    2. a URI for HDFS: `hdfs:///some/path/lines.txt`
+    3. a URI for Amazon's S3 (can also supply credentials inside the URI):
        `s3://my_bucket/lines.txt`, `s3://my_aws_key_id:key_secret@my_bucket/lines.txt`
-    3. HDFS: `hdfs:///some/path/lines.txt`
+    4. an instance of the boto.s3.key.Key class.
 
     Examples::
 
@@ -59,6 +77,12 @@ def smart_open(uri, mode="rb"):
       ...     for line in fin:
       ...         print line
 
+      >>> # you can also use a boto.s3.key.Key instance directly:
+      >>> key = boto.connect_s3().get_bucket("my_bucket").get_key("my_key")
+      >>> with smart_open.smart_open(key) as fin:
+      ...     for line in fin:
+      ...         print line
+
       >>> # stream line-by-line from an HDFS file
       >>> for line in smart_open.smart_open('hdfs:///user/hadoop/my_file.txt'):
       ...    print line
@@ -79,33 +103,58 @@ def smart_open(uri, mode="rb"):
       ...    fout.write("good bye!\n")
 
     """
-    # this method just routes the request to classes handling the specific storage
-    # schemes, depending on the URI protocol in `uri`
-    parsed_uri = ParseUri(uri)
 
-    if parsed_uri.scheme in ("file", ):
-        # local files -- both read & write supported
-        # compression, if any, is determined by the filename extension (.gz, .bz2)
-        return file_smart_open(parsed_uri.uri_path, mode)
+    # validate mode parameter
+    if not isinstance(mode, six.string_types):
+        raise TypeError('mode should be a string')
+    if not mode in ('r', 'rb', 'w', 'wb'):
+        raise NotImplementedError('unknown file mode %s' % mode)
 
-    if mode in ('r', 'rb'):
-        if parsed_uri.scheme in ("s3", "s3n"):
-            return S3OpenRead(parsed_uri)
-        elif parsed_uri.scheme in ("hdfs", ):
-            return HdfsOpenRead(parsed_uri)
-        else:
-            raise NotImplementedError("read mode not supported for %r scheme", parsed_uri.scheme)
-    elif mode in ('w', 'wb'):
-        if parsed_uri.scheme in ("s3", "s3n"):
+    if isinstance(uri, six.string_types):
+        # this method just routes the request to classes handling the specific storage
+        # schemes, depending on the URI protocol in `uri`
+        parsed_uri = ParseUri(uri)
+
+        if parsed_uri.scheme in ("file", ):
+            # local files -- both read & write supported
+            # compression, if any, is determined by the filename extension (.gz, .bz2)
+            return file_smart_open(parsed_uri.uri_path, mode)
+        elif parsed_uri.scheme in ("s3", "s3n"):
             s3_connection = boto.connect_s3(aws_access_key_id=parsed_uri.access_id, aws_secret_access_key=parsed_uri.access_secret)
-            outbucket = s3_connection.get_bucket(parsed_uri.bucket_id)
-            outkey = boto.s3.key.Key(outbucket)
-            outkey.key = parsed_uri.key_id
-            return S3OpenWrite(outbucket, outkey)
+            bucket = s3_connection.get_bucket(parsed_uri.bucket_id)
+            if mode in ('r', 'rb'):
+                key = bucket.get_key(parsed_uri.key_id)
+                if key is None:
+                    raise KeyError(parsed_uri.key_id)
+                return S3OpenRead(key, **kw)
+            else:
+                key = bucket.get_key(parsed_uri.key_id, validate=False)
+                if key is None:
+                    raise KeyError(parsed_uri.key_id)
+                return S3OpenWrite(key, **kw)
+        elif parsed_uri.scheme in ("hdfs", ):
+            if mode in ('r', 'rb'):
+                return HdfsOpenRead(parsed_uri, **kw)
+            else:
+                raise NotImplementedError("write mode not supported for %r scheme", parsed_uri.scheme)
+        elif parsed_uri.scheme in ("webhdfs", ):
+            if mode in ('r', 'rb'):
+                return WebHdfsOpenRead(parsed_uri, **kw)
+            else:
+                return WebHdfsOpenWrite(parsed_uri, **kw)
         else:
-            raise NotImplementedError("write mode not supported for %r scheme", parsed_uri.scheme)
+            raise NotImplementedError("scheme %r is not supported", parsed_uri.scheme)
+    elif isinstance(uri, boto.s3.key.Key):
+        # handle case where we are given an S3 key directly
+        if mode in ('r', 'rb'):
+            return S3OpenRead(uri)
+        elif mode in ('w', 'wb'):
+            return S3OpenWrite(uri)
+    elif hasattr(uri, 'read'):
+        # simply pass-through if already a file-like
+        return uri
     else:
-        raise NotImplementedError("unknown file mode %s" % mode)
+        raise TypeError('don\'t know how to handle uri %s' % repr(uri))
 
 
 class ParseUri(object):
@@ -119,6 +168,8 @@ class ParseUri(object):
       * s3://my_bucket/my_key
       * s3://my_key:my_secret@my_bucket/my_key
       * hdfs:///path/file
+      * hdfs://path/file
+      * webhdfs://host:port/path/file
       * ./local/path/file
       * ./local/path/file.gz
       * file:///home/user/file
@@ -130,14 +181,25 @@ class ParseUri(object):
         Assume `default_scheme` if no scheme given in `uri`.
 
         """
+        if os.name == 'nt':
+            # urlsplit doesn't work on Windows -- it parses the drive as the scheme...
+            if '://' not in uri:
+                # no protocol given => assume a local file
+                uri = 'file://' + uri
         parsed_uri = urlsplit(uri)
         self.scheme = parsed_uri.scheme if parsed_uri.scheme else default_scheme
 
         if self.scheme == "hdfs":
             self.uri_path = parsed_uri.netloc + parsed_uri.path
+            self.uri_path = "/" + self.uri_path.lstrip("/")
 
             if not self.uri_path:
                 raise RuntimeError("invalid HDFS URI: %s" % uri)
+        elif self.scheme == "webhdfs":
+            self.uri_path = parsed_uri.netloc + "/webhdfs/v1" + parsed_uri.path
+
+            if not self.uri_path:
+                raise RuntimeError("invalid WebHDFS URI: %s" % uri)
         elif self.scheme in ("s3", "s3n"):
             self.bucket_id = (parsed_uri.netloc + parsed_uri.path).split('@')
             self.key_id = None
@@ -168,7 +230,7 @@ class ParseUri(object):
             if not self.uri_path:
                 raise RuntimeError("invalid file URI: %s" % uri)
         else:
-            raise NotImplementedError("unknown URI scheme in %r" % uri)
+            raise NotImplementedError("unknown URI scheme %r in %r" % (self.scheme, uri))
 
 
 class S3OpenRead(object):
@@ -176,24 +238,17 @@ class S3OpenRead(object):
     Implement streamed reader from S3, as an iterable & context manager.
 
     """
-    def __init__(self, parsed_uri):
-        if parsed_uri.scheme not in ("s3", "s3n"):
-            raise TypeError("can only process S3 files")
-        self.parsed_uri = parsed_uri
-        s3_connection = boto.connect_s3(
-            aws_access_key_id=parsed_uri.access_id,
-            aws_secret_access_key=parsed_uri.access_secret)
-        self.read_key = s3_connection.get_bucket(parsed_uri.bucket_id).lookup(parsed_uri.key_id)
-        if self.read_key is None:
-            raise KeyError(parsed_uri.key_id)
+    def __init__(self, read_key):
+        if not hasattr(read_key, "bucket") and not hasattr(read_key, "name") and not hasattr(read_key, "read") \
+                and not hasattr(read_key, "close"):
+            raise TypeError("can only process S3 keys")
+        self.read_key = read_key
+        self.line_generator = s3_iter_lines(self.read_key)
 
     def __iter__(self):
-        s3_connection = boto.connect_s3(
-            aws_access_key_id=self.parsed_uri.access_id,
-            aws_secret_access_key=self.parsed_uri.access_secret)
-        key = s3_connection.get_bucket(self.parsed_uri.bucket_id).lookup(self.parsed_uri.key_id)
+        key = self.read_key.bucket.get_key(self.read_key.name)
         if key is None:
-            raise KeyError(self.parsed_uri.key_id)
+            raise KeyError(self.read_key.name)
 
         return s3_iter_lines(key)
 
@@ -229,6 +284,12 @@ class S3OpenRead(object):
     def __exit__(self, type, value, traceback):
         self.read_key.close()
 
+    def __str__(self):
+        return "%s<key: %s>" % (
+            self.__class__.__name__, self.read_key
+        )
+
+
 
 class HdfsOpenRead(object):
     """
@@ -241,7 +302,7 @@ class HdfsOpenRead(object):
         self.parsed_uri = parsed_uri
 
     def __iter__(self):
-        hdfs = subprocess.Popen(["hadoop", "fs", "-cat", self.parsed_uri.uri_path], stdout=subprocess.PIPE)
+        hdfs = subprocess.Popen(["hdfs", "dfs", "-cat", self.parsed_uri.uri_path], stdout=subprocess.PIPE)
         return hdfs.stdout
 
     def read(self, size=None):
@@ -257,6 +318,61 @@ class HdfsOpenRead(object):
         pass
 
 
+class WebHdfsOpenRead(object):
+    """
+    Implement streamed reader from WebHDFS, as an iterable & context manager.
+    NOTE: it does not support kerberos authentication yet
+
+    """
+    def __init__(self, parsed_uri):
+        if parsed_uri.scheme not in ("webhdfs"):
+            raise TypeError("can only process WebHDFS files")
+        self.parsed_uri = parsed_uri
+        self.offset = 0
+
+    def __iter__(self):
+        payload = {"op": "OPEN"}
+        response = requests.get("http://" + self.parsed_uri.uri_path, params=payload, stream=True)
+        return response.iter_lines()
+
+    def read(self, size=None):
+        """
+        Read the specific number of bytes from the file
+
+        Note read() and line iteration (`for line in self: ...`) each have their
+        own file position, so they are independent. Doing a `read` will not affect
+        the line iteration, and vice versa.
+        """
+        if not size or size < 0:
+            payload = {"op": "OPEN", "offset": self.offset}
+            self.offset = 0
+        else:
+            payload = {"op": "OPEN", "offset": self.offset, "length": size}
+            self.offset = self.offset + size
+        response = requests.get("http://" + self.parsed_uri.uri_path, params=payload, stream=True)
+        return response.content
+
+    def seek(self, offset, whence=0):
+        """
+        Seek to the specified position.
+
+        Only seeking to the beginning (offset=0) supported for now.
+
+        """
+        if whence == 0 and offset == 0:
+            self.offset = 0
+        elif whence == 0:
+            self.offset = offset
+        else:
+            raise NotImplementedError("operations with whence not implemented yet")
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, type, value, traceback):
+        pass
+
+
 def make_closing(base, **attrs):
     """
     Add support for `with Base(attrs) as fout:` to the base class if it's missing.
@@ -301,14 +417,16 @@ class S3OpenWrite(object):
     Context manager for writing into S3 files.
 
     """
-    def __init__(self, outbucket, outkey, min_part_size=S3_MIN_PART_SIZE):
+    def __init__(self, outkey, min_part_size=S3_MIN_PART_SIZE, **kw):
         """
         Streamed input is uploaded in chunks, as soon as `min_part_size` bytes are
         accumulated (50MB by default). The minimum chunk size allowed by AWS S3
         is 5MB.
 
         """
-        self.outbucket = outbucket
+        if not hasattr(outkey, "bucket") and not hasattr(outkey, "name"):
+            raise TypeError("can only process S3 keys")
+
         self.outkey = outkey
         self.min_part_size = min_part_size
 
@@ -316,7 +434,7 @@ class S3OpenWrite(object):
             logger.warning("S3 requires minimum part size >= 5MB; multipart upload may fail")
 
         # initialize mulitpart upload
-        self.mp = self.outbucket.initiate_multipart_upload(self.outkey)
+        self.mp = self.outkey.bucket.initiate_multipart_upload(self.outkey, **kw)
 
         # initialize stats
         self.lines = []
@@ -324,6 +442,11 @@ class S3OpenWrite(object):
         self.chunk_bytes = 0
         self.parts = 0
 
+    def __str__(self):
+        return "%s<key: %s, min_part_size: %s>" % (
+            self.__class__.__name__, self.outkey, self.min_part_size,
+            )
+
     def write(self, b):
         """
         Write the given bytes (binary string) into the S3 file from constructor.
@@ -368,19 +491,105 @@ class S3OpenWrite(object):
             # when the input is completely empty => abort the upload, no file created
             # TODO: or create the empty file some other way?
             logger.info("empty input, ignoring multipart upload")
-            self.outbucket.cancel_multipart_upload(self.mp.key_name, self.mp.id)
+            self.outkey.bucket.cancel_multipart_upload(self.mp.key_name, self.mp.id)
+
+    def __enter__(self):
+        return self
+
+    def _termination_error(self):
+        logger.exception("encountered error while terminating multipart upload; attempting cancel")
+        self.outkey.bucket.cancel_multipart_upload(self.mp.key_name, self.mp.id)
+        logger.info("cancel completed")
+
+    def __exit__(self, type, value, traceback):
+        if type is not None:
+            self._termination_error()
+            return False
+
+        try:
+            self.close()
+        except:
+            self._termination_error()
+            raise
+
+
+class WebHdfsOpenWrite(object):
+    """
+    Context manager for writing into webhdfs files
+
+    """
+    def __init__(self, parsed_uri, min_part_size=WEBHDFS_MIN_PART_SIZE):
+        if parsed_uri.scheme not in ("webhdfs"):
+            raise TypeError("can only process WebHDFS files")
+        self.parsed_uri = parsed_uri
+        self.closed = False
+        self.min_part_size = min_part_size
+        # creating empty file first
+        payload = {"op": "CREATE", "overwrite": True}
+        init_response = requests.put("http://" + self.parsed_uri.uri_path, params=payload, allow_redirects=False)
+        if not init_response.status_code == httplib.TEMPORARY_REDIRECT:
+            raise WebHdfsException(str(init_response.status_code) + "\n" + init_response.content)
+        uri = init_response.headers['location']
+        response = requests.put(uri, data="", headers={'content-type': 'application/octet-stream'})
+        if not response.status_code == httplib.CREATED:
+            raise WebHdfsException(str(response.status_code) + "\n" + response.content)
+        self.lines = []
+        self.parts = 0
+        self.chunk_bytes = 0
+        self.total_size = 0
+
+    def upload(self, data):
+        payload = {"op": "APPEND"}
+        init_response = requests.post("http://" + self.parsed_uri.uri_path, params=payload, allow_redirects=False)
+        if not init_response.status_code == httplib.TEMPORARY_REDIRECT:
+            raise WebHdfsException(str(init_response.status_code) + "\n" + init_response.content)
+        uri = init_response.headers['location']
+        response = requests.post(uri, data=data, headers={'content-type': 'application/octet-stream'})
+        if not response.status_code == httplib.OK:
+            raise WebHdfsException(str(response.status_code) + "\n" + response.content)
+
+    def write(self, b):
+        """
+        Write the given bytes (binary string) into the WebHDFS file from constructor.
+
+        """
+        if self.closed:
+            raise ValueError("I/O operation on closed file")
+        if isinstance(b, six.text_type):
+            # not part of API: also accept unicode => encode it as utf8
+            b = b.encode('utf8')
+
+        if not isinstance(b, six.binary_type):
+            raise TypeError("input must be a binary string")
+
+        self.lines.append(b)
+        self.chunk_bytes += len(b)
+        self.total_size += len(b)
+
+        if self.chunk_bytes >= self.min_part_size:
+            buff = b"".join(self.lines)
+            logger.info("uploading part #%i, %i bytes (total %.3fGB)" % (self.parts, len(buff), self.total_size / 1024.0 ** 3))
+            self.upload(buff)
+            logger.debug("upload of part #%i finished" % self.parts)
+            self.parts += 1
+            self.lines, self.chunk_bytes = [], 0
+
+    def seek(self, offset, whence=None):
+        raise NotImplementedError("seek() not implemented yet")
+
+    def close(self):
+        buff = b"".join(self.lines)
+        if buff:
+            logger.info("uploading last part #%i, %i bytes (total %.3fGB)" % (self.parts, len(buff), self.total_size / 1024.0 ** 3))
+            self.upload(buff)
+            logger.debug("upload of last part #%i finished" % self.parts)
+        self.closed = True
 
     def __enter__(self):
         return self
 
     def __exit__(self, type, value, traceback):
-        try:
-            self.close()
-        except:
-            logger.exception("encountered error while terminating multipart upload; attempting cancel")
-            self.outbucket.cancel_multipart_upload(self.mp.key_name, self.mp.id)
-            logger.info("cancel completed")
-            raise
+        self.close()
 
 
 def s3_iter_bucket_process_key(key):
@@ -395,7 +604,7 @@ def s3_iter_bucket_process_key(key):
 def s3_iter_bucket(bucket, prefix='', accept_key=lambda key: True, key_limit=None, workers=16):
     """
     Iterate and download all S3 files under `bucket/prefix`, yielding out
-    `(key name, key content)` 2-tuples (generator).
+    `(key, key content)` 2-tuples (generator).
 
     `accept_key` is a function that accepts a key name (unicode string) and
     returns True/False, signalling whether the given key should be downloaded out or
@@ -404,7 +613,8 @@ def s3_iter_bucket(bucket, prefix='', accept_key=lambda key: True, key_limit=Non
     If `key_limit` is given, stop after yielding out that many results.
 
     The keys are processed in parallel, using `workers` processes (default: 16),
-    to speed up downloads greatly.
+    to speed up downloads greatly. If multiprocessing is not available, thus
+    NO_MULTIPROCESSING is True, this parameter will be ignored.
 
     Example::
 
@@ -419,25 +629,32 @@ def s3_iter_bucket(bucket, prefix='', accept_key=lambda key: True, key_limit=Non
       ...     print key, len(content)
 
     """
-    logger.info("iterating over keys from %s with %i workers" % (bucket, workers))
-
     total_size, key_no = 0, -1
     keys = (key for key in bucket.list(prefix=prefix) if accept_key(key.name))
 
-    pool = multiprocessing.pool.Pool(processes=workers)
-    for key_no, (key, content) in enumerate(pool.imap_unordered(s3_iter_bucket_process_key, keys)):
+    if NO_MULTIPROCESSING:
+        logger.info("iterating over keys from %s without multiprocessing" % bucket)
+        iterator = imap(s3_iter_bucket_process_key, keys)
+    else:
+        logger.info("iterating over keys from %s with %i workers" % (bucket, workers))
+        pool = multiprocessing.pool.Pool(processes=workers)
+        iterator = pool.imap_unordered(s3_iter_bucket_process_key, keys)
+
+    for key_no, (key, content) in enumerate(iterator):
         if key_no % 1000 == 0:
             logger.info("yielding key #%i: %s, size %i (total %.1fMB)" %
                 (key_no, key, len(content), total_size / 1024.0 ** 2))
 
-        yield key.name, content
+        yield key, content
         key.close()
         total_size += len(content)
 
         if key_limit is not None and key_no + 1 >= key_limit:
             # we were asked to output only a limited number of keys => we're done
             break
-    pool.terminate()
+
+    if not NO_MULTIPROCESSING:
+        pool.terminate()
 
     logger.info("processed %i keys, total size %i" % (key_no + 1, total_size))
 
@@ -473,3 +690,9 @@ def s3_iter_lines(key):
     # process the last line, too
     if buf:
         yield buf
+
+
+class WebHdfsException(Exception):
+    def __init__(self, msg=str()):
+        self.msg = msg
+        super(WebHdfsException, self).__init__(self.msg)
